' Example to query PIDs for the battery ECU
' and convert the received data into engineering values
' JDH 10/8/12


Dim ok
Dim txID : Dim txData(8) : Dim txLen
Dim rxID : Dim rxData(8)

txData(0) = 0 : txData(1) = 0 : txData(2) = 0 : txData(3) = 0
txData(4) = 0 : txData(5) = 0 : txData(6) = 0 : txData(7) = 0
txLen = 8 ' always transmit 8 bytes even though payload may be less

txID = &H7E3    ' Battery ECU module
rxID = txID + 8 ' id of reply is 8 higher than module number

' Check the "ok" result, if it fails print a message and exit
Sub checkOK(okay, failed$, succeeded$, xit)
  If (okay=0) Then
    Print failed$
    If (xit=1) Then Exit
  Else
    Print succeeded$
  EndIf
End Sub

' print formatted hex numbers
Function toHex$(val)
  toHex$=""
  If (val<16) Then toHex$="0"
  toHex$ = toHex$ + Hex$(val)
End Function

' print timer : id : len : data
Sub PrintRawData
    Print Timer ": " Hex$(rxId) " : " rxLen " : " toHex$(rxData(0)) " ";
    Print toHex$(rxData(1)) " " toHex$(rxData(2)) " " toHex$(rxData(3)) " ";
    Print toHex$(rxData(4)) " " toHex$(rxData(5)) " " toHex$(rxData(6)) " ";
    Print toHex$(rxData(7))
End Sub

' make the PID request
Sub SendModeAndPID(mode, pid)
  Local txOK
  txData(0) = 3
  txData(1) = mode
  txData(2) = pid
  CAN TX 0,txID,0,txLen,txData(0),txOk
  checkOK(txOK, "FAILED TO SEND PID", "SENT PID " + Hex$(pid) + " TO " + Hex$(txId), 0)
End Sub

' Send the acknowledgement after receiving frame
Sub SendReadyForMore()
  Local txOK
  txData(0) = &H30
  txData(1) = 0
  txData(2) = 0
  CAN TX 0,txID,0,txLen,txData(0),txOk
  checkOK(txOK, "FAILED TO SEND PID", "SENT 0x30 TO " + Hex$(txId), 0)
End Sub

' Chains together the initial PID request and the acknowledgement
Sub RequestPID(mode, pid)
  ' Note, we're "cheating" here and simply waiting 10msec and sending the
  ' acknowlegement.  Strictly speaking we should wait for the response
  ' header first (frame 0x10).  This works because in the CAN setup (below)
  ' we've setup a big enough buffer to receive all the frames from the
  ' longest PID.
  SendModeAndPID(mode, pid)
  Pause 10
  SendReadyForMore
End Sub

Sub DisplayD0
  ' retrieve the data (see DisplayCE for commentary)
  RequestPID &H21,&HD0
  Do
    CAN RX 1, rxData(0), ok
  Loop Until (ok=1)
  PrintRawData
  Local b(rxData(1)+15)
  Local j
  Local i
  j=0
  For i=4 To 7
    b(j)=rxData(i)
    j=j+1
  Next i
  Do
    CAN RX 1, rxData(0), ok
    If (ok=1) Then
      PrintRawData
      For i=1 To 7
        b(j)=rxData(i)
        j=j+1
      Next i
    EndIf
  Loop Until (ok=0)

  ' Convert to engineering units (see DisplayCE for commentary)
  Print "Block Count         = " b(0)
  Print "Time in LOW         = " (256*b(1)+b(2))
  Print "Time in DC Inhibit  = " (256*(b3)+b(4))
  Print "Time in HIGH        = " (256*(b5)+b(6))
  Print "Time in HOT         = " (256*(b7)+b(8))
  Print "Lowest Block        = " b(9)
  Print "Lowest Voltage      = " (2.56*b(10)+0.1*b(11)-327.68)
  Print "Highest Block       = " b(12)
  Print "Highest Voltage     = " (2.56*b(13)+0.1*b(14)-327.68)
  For i=15 To 28
    Print "Block " toHex$(i-14) " Resistance = " (0.001*b(i)) " Ohms"
  Next i
End Sub

Sub DisplayCF
  ' retrieve the data (see DisplayCE for commentary)
  RequestPID &H21,&HCF
  Do
    CAN RX 1, rxData(0), ok
  Loop Until (ok=1)
  PrintRawData
  Local b(rxData(1)+15)
  Local j
  Local i
  j=0
  For i=4 To 7
    b(j)=rxData(i)
    j=j+1
  Next i
  Do
    CAN RX 1, rxData(0), ok
    If (ok=1) Then
      PrintRawData
      For i=1 To 7
        b(j)=rxData(i)
        j=j+1
      Next i
    EndIf
  Loop Until (ok=0)

  ' Convert to engineering units (see DisplayCE for commentary)
  Print "Air Intake Temp   = " (4.608*b(0)+0.018*b(1)-557.824) " F"
  Print "Fan Motor Voltage = " (0.2*b(2) - 25.6) " V"
  Print "Aux Batt Voltage  = " (0.2*b(3) - 25.6) " V"
  Print "Battery Charge    = " (b(4) - 64)
  Print "Battery Discharge = " (b(5) - 64)
  Print "Delta SOC         = " (0.01 * b(6)) " %"
  Print "Fan Speed         = " b(8)
  Print "Batt Temp 1       = " (4.608*b(10)+0.018*b(11)-557.824) " F"
  Print "Batt Temp 2       = " (4.608*b(12)+0.018*b(13)-557.824) " F"
  Print "Batt Temp 3       = " (4.608*b(14)+0.018*b(15)-557.824) " F"
End Sub

Sub DisplayCE
  ' send the request
  RequestPID &H21,&HCE

  ' retrieve the first frame
  Do
    CAN RX 1, rxData(0), ok
  Loop Until (ok=1)
  PrintRawData
  ' create an array big enough to hold all the data
  ' the +15 is because there always seems to be 1 more frame than expected
  ' and if the amount of data is 1 larger than a multiple of 8 we need 7
  ' bytes for the remainder of that frame + 8 bytes for the "extra" frame
  Local buffer(rxData(1)+15)

  ' Copy the data from the first frame into the buffer
  Local i
  Local j
  j=0
  For i=4 To 7
    buffer(j)=rxData(i)
    j=j+1
  Next i

  ' now process the remaining frames
  Do
    CAN RX 1, rxData(0), ok
    If (ok=1) Then
      PrintRawData
      For i=1 To 7
        buffer(j)=rxData(i)
        j=j+1
      Next i
    EndIf
  Loop Until (ok=0)

  ' We have the raw data, now convert it to engineering values
  ' These calculations are from USBSeaWolf's spreadsheet available on the
  ' PriusChat.com forum.
  Print "SOC      = " (0.5 * buffer(0)) " %"
  Print "Current  = " (2.56*buffer(1)+0.1*buffer(2)-327.68) " A"
  j=3
  For i=1 To 14
    Print "Block " toHex$(i) " = " (2.56*buffer(j)+0.1*buffer(j+1)-327.68) " V"
    j=j+2
  Next i
End Sub


''''''''''''''''''''''''''
' START HERE...

Cls

CAN CONFIG ok                : checkOK(ok, "CAN CONFIG FAILED", "", 1)
CAN SETSPEED 500000, ok      : checkOK(ok, "CAN SETSPEED FAILED", "", 1)
CAN ADDTXCHNL 0,1,ok         : checkOK(ok, "CAN ADDTXCHNL FAILED", "", 1)
CAN ADDRXCHNL 1,rxId,0,10,ok : checkOK(ok, "CAN ADDRXCHNL FAILED", "", 1)
CAN ENABLE ok                : checkOK(ok, "CAN ENABLE FAILED", "", 1)
CAN PRINTCONFIG

' Note the above configuration for channel 1
' (a) we're filtering on the response ID (which is the moduleID + 8)
' (b) we create a 10 record buffer this allows us to capture all the response
'     frames in one go

Print "q:quit, t:0xCE, u:0xCF, v:0xD0 send pid to " Hex$(txID)

Timer = 0
Do
  k$ = Inkey$
  If (k$ = "q") Then Exit
  If (k$ = "t") Then DisplayCE
  If (k$ = "u") Then DisplayCF
  If (k$ = "v") Then DisplayD0
  CAN RX 1, rxData(0), ok
  If (ok=1) Then PrintRawData
Loop

CAN FREE

End
